from oml.losses.arcface import ArcFaceLoss, ArcFaceLossWithMLP
from oml.losses.surrogate_precision import SurrogatePrecision
from oml.losses.triplet import TripletLoss, TripletLossPlain, TripletLossWithMiner
